import config
import util

from datetime import datetime
from twisted.python import log

def write_message(message, user, channel):
    for user_name, client in config.irc_users.iteritems():
        #assume *messages* are ACTION and translate accordingly
        if message.startswith('*') and message.endswith('*'):
            # I don't get why client.action doesn't do this in Twisted - mim
            client.privmsg(user, '#%s' % channel, '\x01ACTION %s\x01' % message.strip('*'))
            log.msg('Translating action to %s on %s: %s' % (user_name, channel, message.encode('ascii', 'ignore')))
        else:
            client.privmsg(user, '#%s' % channel, message)
            log.msg('Writing to %s on %s: %s' % (user_name, channel, message.encode('ascii', 'ignore')))

def write_part_message(user, channel, reason=None):
    for user_name, client in config.irc_users.iteritems():
        client.part('%s!campfire@luckystrike' % user, channel, reason)

def write_join_message(user, channel):
    for user_name, client in config.irc_users.iteritems():
        client.join('%s!campfire@luckystrike' % user, channel)

def route_incoming_message(message):
    """
    Take incoming message from Campfire, sort by type and deliver to IRC
    """
    if message['user_id'] is not None:
        user = config.campfire.user(message['user_id'])['user']
    else:
        user = None

    channel = '#' + config.rooms[message['room_id']]['channel']

    # Update last message seen
    config.rooms[message['room_id']]['heartbeat'] = datetime.utcnow()

    # Don't write messages that I've sent, or that aren't Text
    if message['type'] == 'TextMessage' and config.campfire.me()['id'] != message['user_id']:
        write_message(message['body'], util.campNameToString(user['name']),
                config.rooms[message['room_id']]['channel'])
    elif message['type'] == 'EnterMessage':
        log.msg('EnterMessage %s joined %s' % (message['user_id'], message['room_id']))
        nick = util.campNameToString(config.campfire.user(message['user_id'])['user']['name'])
        write_join_message(nick, channel)
    elif message['type'] == 'KickMessage':
        log.msg('KickMessage %s left %s' % (message['user_id'], message['room_id']))
        nick = util.campNameToString(config.campfire.user(message['user_id'])['user']['name'])
        write_part_message(nick, channel, 'Timed out')
    elif message['type'] == 'LeaveMessage':
        log.msg('LeaveMessage %s left %s' % (message['user_id'], message['room_id']))
        nick = util.campNameToString(config.campfire.user(message['user_id'])['user']['name'])
        write_part_message(nick, channel, 'Left channel')
    elif message['type'] == 'PasteMessage':
        # Write first 5 lines of paste
        write_message('Paste: https://%s.campfirenow.com/room/%s/paste/%s' % (config.configuration['domain'], message['room_id'], message['id']), util.campNameToString(user['name']), config.rooms[message['room_id']]['channel'])
        lines = message['body'].splitlines()
        for line in lines[:5]:
            write_message(line, util.campNameToString(user['name']), config.rooms[message['room_id']]['channel'])
        if len(lines) > 5:
            write_message('... Paste too long ...', util.campNameToString(user['name']), config.rooms[message['room_id']]['channel'])
    elif message['type'] == 'SoundMessage':
        write_message('Played a sound: %s' % message['body'], util.campNameToString(user['name']), config.rooms[message['room_id']]['channel'])
        log.msg('SoundMessage: %s' % message['body'])
    elif message['type'] == 'TweetMessage':
        write_message(message['body'], util.campNameToString(user['name']), config.rooms[message['room_id']]['channel'])
    elif message['type'] == 'TimestampMessage':
        config.rooms[message['room_id']]['heartbeat'] =  datetime.strptime(
                message['created_at'][:-6], '%Y/%m/%d %H:%M:%S') 
        if config.args.debug:
            log.msg('Timestamp received: %s' % message['created_at'])
    elif message['type'] == 'UploadMessage':
        write_message('Uploaded: https://%s.campfirenow.com/room/%s/uploads/%s' % (config.configuration['domain'], message['room_id'], message['body']), util.campNameToString(user['name']), config.rooms[message['room_id']]['channel'])
        log.msg('UploadMessage: %s' % message['body'])
    elif message['type'] == 'TopicChangeMessage':
        log.msg('TopicChangeMessage: %s' % message['body'])
        group = config.irc_realm.groups[config.rooms[message['room_id']]['channel']]
        group.setMetadata({'topic': message['body']})
    elif message['type'] == 'AllowGuestsMessage':
        write_message('Has allowed Guests', util.campNameToString(user['name']), config.rooms[message['room_id']]['channel'])
        log.msg('AllowGuestsMessage: %s' % message['body'])
    elif message['type'] == 'DisallowGuestsMessage':
        write_message('Has Disallowed Guests', util.campNameToString(user['name']), config.rooms[message['room_id']]['channel'])
        log.msg('DisallowGuestsMessage: %s' % message['body'])
    elif config.campfire.me()['id'] != message['user_id']:
        log.err('Unknown message type received: %s' % message)



import argparse
import json
from os import path

import pinder

from twisted.words import service

rooms = {}
irc_users = {}

if path.exists(path.expanduser('~') + '/.luckystrike/config.json'):
    with open(path.expanduser('~') + '/.luckystrike/config.json') as config_file:
        configuration = dict(json.loads(config_file.read()))
else:
    with open('config.json') as config_file:
        configuration = dict(json.loads(config_file.read()))

users = configuration['users']

parser = argparse.ArgumentParser(description='Campfire to IRC Proxy')
parser.add_argument(
        '-d', '--debug', 
        action='store_true', 
        help='increase debug level'
)
parser.add_argument(
        '-s', '--setup_config', 
        action='store_true', 
        help='generate config.json'
)
args = parser.parse_args()

# connect to Campfire
campfire = pinder.Campfire(configuration['domain'], configuration['api_key'])
# Initialize the Cred authentication system used by the IRC server.
irc_realm = service.InMemoryWordsRealm('LuckyStrike')



from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.python import log

import httpstream

class MessageReceiver(httpstream.MessageReceiver):

    """
    Receive Message from stream and callback the provided method else errback
    """

    def __init__(self, user_callback, user_errback, url):
        self.user_callback = user_callback
        self.user_errback = user_errback
        self.url = url

    def connectionMade(self):
        log.msg('started receiving messages from %s' % self.url)

    def connectionFailed(self, why):
        log.err('cannot connect to %s: %s' % (self.url, why))

    def messageReceived(self, message):
        d = Deferred()
        d.addCallback(self.user_callback)
        d.addErrback(self.user_errback)

        d.callback(message)

def listen(username, room_id, callback, errback):
    """
    Start Campfire live stream
    """
    url = 'https://streaming.campfirenow.com/room/%s/live.json' % room_id
    stream = httpstream.stream(reactor, url, MessageReceiver(callback, errback, url), username=username, password='X')

    return stream



#!/usr/bin/env python
# coding=utf-8

import json
import os
import sys

from collections import defaultdict

from twisted.conch import manhole, manhole_ssh
from twisted.cred import checkers, portal
from twisted.internet import reactor, ssl
from twisted.internet import task
from twisted.python import log
from twisted.words import service

from luckystrike import util
from luckystrike import config
from luckystrike.user import LuckyStrikeIRCFactory

def setup_config():
    """
    Easier walkthrough for user to setup .luckystrike/config
    """
    print 'Please input domain, without tld, for example.com, input example'
    domain = raw_input('Domain: ')
    user = raw_input('IRC nickname: ')
    password = util.generate_password()
    api_key = raw_input('Campfire API key: ')

    d = defaultdict()
    d['domain'] = domain
    d['users'] = {user : password}
    d['api_key'] = api_key

    print 'Your username:%s, and password: %s' % (user, password)

    directory = os.path.expanduser('~') + '/.luckystrike'
    if not os.path.exists(directory):
        os.makedirs(directory)

    print 'Writing config to %sconfig.json' % directory

    with open(directory + '/config.json', 'w') as config_file:
        json.dump(d, config_file)

def getManholeFactory(namespace, **passwords):
    """
    Return a Manhole SSH factory for debug purposes
    """
    realm = manhole_ssh.TerminalRealm()

    def getManhole(_):
        return manhole.Manhole(namespace)

    realm.chainedProtocolFactory.protocolFactory = getManhole
    p = portal.Portal(realm)
    p.registerChecker(checkers.InMemoryUsernamePasswordDatabaseDontUse(**passwords))
    f = manhole_ssh.ConchFactory(p)

    return f

def watchdog():
    """
    Track last message received per room, call periodically in a reactor loop
    """
    for room in config.rooms.itervalues():
        if room['streaming']:
            log.msg('Last timestamp from %s is %s' % (room['channel'],
                room['heartbeat']))

if __name__ == '__main__':
    if config.args.setup_config:
        setup_config()
        sys.exit(0)

    log.startLogging(sys.stdout)

    try:
        # Store pid file first thing
        pidfile = config.configuration.get('pidfile', None)
        if pidfile is not None:
            with open(pidfile, 'w+') as pf:
                pf.write("%s\n" % str(os.getpid()))

        for room in config.campfire.rooms():
            config.rooms[room['id']] = room
            config.rooms[room['id']]['channel'] = util.campNameToString(room['name'])
            config.rooms[room['id']]['stream'] = None
            config.rooms[room['id']]['streaming'] = False
            config.rooms[room['id']]['heartbeat'] = None

            log.msg('Adding %s to IRC as %s' % (room['name'],
                config.rooms[room['id']]['channel']))

            config.irc_realm.addGroup(service.Group(config.rooms[room['id']]['channel']))

        user_db = checkers.InMemoryUsernamePasswordDatabaseDontUse(**config.users)
        irc_portal = portal.Portal(config.irc_realm, [user_db])

        # Start IRC and Manhole
        reactor.listenTCP(int(config.configuration.get('port', 6667)),
                LuckyStrikeIRCFactory(config.irc_realm, irc_portal))

        if config.args.debug:
            admin_password = util.generate_password()
            log.msg('Staring ManHole with admin password: %s' % admin_password)
            reactor.listenTCP(
                    int(config.configuration.get('manhole_port', 2222)),
                    getManholeFactory(globals(), admin=admin_password),
                    interface='127.0.0.1'
            )

        if (os.path.exists(config.configuration.get('ssl_crt', '')) and
            os.path.exists(config.configuration.get('ssl_key', ''))):

            reactor.listenSSL(int(config.configuration.get('ssl_port', 6697)),
                    LuckyStrikeIRCFactory(config.irc_realm, irc_portal),
                    ssl.DefaultOpenSSLContextFactory(
                        config.configuration['ssl_key'],
                        config.configuration['ssl_crt'])
                    )

        if config.args.debug:
            t = task.LoopingCall(watchdog)
            t.start(240)

        reactor.run()

    except:
        log.err()
    finally:
        if pidfile is not None:
            os.unlink(pidfile)



import config
import random
import re
import string
import sys
import traceback

from twisted.python import log

def generate_password(length = 12):
    """
    Generate a random password, ASCII letters + digits only, default length 12
    """
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(length))

def campNameToString(name):
    """
    Translate Campfire String to one renderable on IRC for nicknames, or channels
    """
    return re.sub('\s+', '_', name).lower()

def channel_to_room(channel):
    """
    Given an IRC channel, return a Campfire room object
    """
    for room_id, room in config.rooms.iteritems():
        if room['channel'] == channel:
            return room

def error(e):
    """
    Take an exception and render it to an IRC notice message for all clients
    """
    exc_type, exc_value, exc_traceback = sys.exc_info()
    log.err(e)
    for user_name, client in config.irc_users.iteritems():
        for line in traceback.format_exception(exc_type, exc_value, exc_traceback):
            client.notice('LuckyStrike', user_name, line)



import re

from twisted.internet import defer
from twisted.python import log
from twisted.words import service
from twisted.words.protocols import irc

from luckystrike import util
from luckystrike import stream
from luckystrike import config
from luckystrike import route

def replace_usernames(room, message):
    """
    Search message for potential usernames, return message with expanded name
    """
    users = {}
    replace = []

    #consider caching this, but if we error out, just don't worry about
    #replacing usernames
    try:
        for user in room._get()['room']['users']:
            users[util.campNameToString(user['name'])] = user['name']
    except:
        log.err()
        return message

    # find all words that match usernames
    for m in message.split():
        a = [(k, v) for (k, v) in users.iteritems() if re.sub(r'\W+', '', m) == k]
        for i in a:
            replace.append(i)

    # search and replace all irc to campfire names
    if len(replace) > 0:
        log.msg('Found usernames to replace in message :%s' % replace)
        message_replacement = message
        for r in replace:
            p = re.compile(r[0])
            message_replacement = p.sub(r[1], message_replacement)

        log.msg('Transforming %s into: %s' % (message, message_replacement))

        return message_replacement
    else:
        return message

class LuckyStrikeIRCUser(service.IRCUser):

    """
    Override Twisted's IRC User
    """

    def connectionMade(self):
        service.IRCUser.connectionMade(self)
        log.msg('User connected from %s' % self.hostname)

    def _cbLogin(self, (iface, avatar, logout)):
        service.IRCUser._cbLogin(self, (iface, avatar, logout))
        log.msg('User authenticated as: %s' % self.avatar.name)
        config.irc_users[self.avatar.name] = self

    def channelToRoom(self, channel):
        room_info = util.channel_to_room(channel.strip('#'))
        room = config.campfire.find_room_by_name(room_info['name'])

        return room

    def connectionLost(self, reason):
        log.msg('User disconnected: %s', reason)
        if self.avatar is not None:
            del config.irc_users[self.avatar.name]
        service.IRCUser.connectionLost(self, reason)

    def names(self, user, channel, names):
        log.msg('names called: %s, %s, %s' % (user, channel, names))

        room = self.channelToRoom(channel)

        names = []
        users = room._get()['room']['users']
        for camp_user in users:
            names.append(util.campNameToString(camp_user['name']))

        service.IRCUser.names(self, user, channel, names)

    def who(self, user, channel, memberInfo):
        #[(u'mmichie', 'LuckyStrike', 'LuckyStrike', u'mmichie', 'H', 0, u'mmichie')]
        log.msg('who called: %s, %s, %s' % (user, channel, memberInfo))
        room = self.channelToRoom(channel)

        users = room._get()['room']['users']
        for camp_user in users:
            name = util.campNameToString(camp_user['name'])
            memberInfo.append((name, 'LuckyStrike', 'LuckyStrike', name, 'H', 0, name))

        service.IRCUser.who(self, user, channel, memberInfo)

    def irc_JOIN(self, prefix, params):
        for channel in params[0].split(','):
            room_info = util.channel_to_room(channel.strip('#'))
            group = config.irc_realm.groups[util.campNameToString(room_info['name'])]
            group.setMetadata({'topic': room_info['topic']})

            service.IRCUser.irc_JOIN(self, prefix, [channel])
            log.msg('Joined channel: %s, %s' % (prefix, channel))

            # Join Campfire room
            log.msg('Starting to stream: %s' % room_info['name'])
            room = config.campfire.find_room_by_name(room_info['name'])
            room.join()

            if config.rooms[room.id]['stream'] is None:
                username, password = room._connector.get_credentials()
                config.rooms[room.id]['stream'] = stream.listen(username, room.id,
                    route.route_incoming_message, util.error)

    def irc_TOPIC(self, prefix, params):
        service.IRCUser.irc_TOPIC(self, prefix, params)
        channel = params[0].strip('#')
        topic = params[1]
        room = self.channelToRoom(channel)
        log.msg('Setting topic: %s on %s' % (room.name, topic))
        room.update(room.name, topic)

    def irc_PRIVMSG(self, prefix, params):
        service.IRCUser.irc_PRIVMSG(self, prefix, params)
        if params[0].startswith('#'):
            room = self.channelToRoom(params[0])
            message = replace_usernames(room, params[1])

            # Handle /me
            if message.startswith('\x01ACTION'):
                before = message
                message = message.strip('\x01')
                message = '*' + ' '.join(message.split()[1:]) + '*'
                log.msg('ACTION Translated from %s to %s' % (before, message))

            log.msg('Speaking to %s: %s' % (room.name, message.decode('ascii', 'ignore')))
            try:
                room.speak(message)
            except Exception as exception:
                util.error(exception)

    def irc_PART(self, prefix, params):
        service.IRCUser.irc_PART(self, prefix, params)
        log.msg('Left channel: %s, %s' % (prefix, params))

        # Leave Campfire room
        room_info = util.channel_to_room(params[0].strip('#'))
        log.msg('Stopping stream to : %s' % room_info['name'])
        room = config.campfire.find_room_by_name(room_info['name'])
        room.leave()
        try:
            factory, connection = config.rooms[room.id]['stream']
            factory.stopTrying()
            connection.disconnect()
            config.rooms[room.id]['stream'] = None
        except:
            log.err()

    def irc_ISON(self, prefix, params):
        pass

    def irc_CAP(self, prefix, params):
        pass

    def irc_AWAY(self, prefix, params):
        pass

    def irc_LIST(self, prefix, params):
        """List query

        Return information about the indicated channels, or about all
        channels if none are specified.

        Parameters: [ <channel> *( "," <channel> ) [ <target> ] ]
        """
        #<< list #python
        #>> :orwell.freenode.net 321 exarkun Channel :Users  Name
        #>> :orwell.freenode.net 322 exarkun #python 358 :The Python programming language
        #>> :orwell.freenode.net 323 exarkun :End of /LIST
        if params:
            # Return information about indicated channels
            try:
                channels = params[0].decode(self.encoding).split(',')
            except UnicodeDecodeError:
                self.sendMessage(
                    irc.ERR_NOSUCHCHANNEL, params[0],
                    ":No such channel (could not decode your unicode!)")
                return

            groups = []
            for ch in channels:
                if ch.startswith('#'):
                    ch = ch[1:]
                groups.append(self.realm.lookupGroup(ch))

            groups = defer.DeferredList(groups, consumeErrors=True)
            groups.addCallback(lambda gs: [r for (s, r) in gs if s])
        else:
            # Return information about all channels
            groups = self.realm.itergroups()

        def cbGroups(groups):
            def gotSize(size, group):
                return '#'+group.name, size, group.meta.get('topic')
            d = defer.DeferredList([
                group.size().addCallback(gotSize, group) for group in groups])
            d.addCallback(lambda results: self.list([r for (s, r) in results if s]))
            return d
        groups.addCallback(cbGroups)

class LuckyStrikeIRCFactory(service.IRCFactory):
    protocol = LuckyStrikeIRCUser



# coding: utf-8
#
# Copyright 2009 Alexandre Fiori
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

__author__ = "Thomas Rabaix, Alexandre Fiori"
__version__ = "0.0.1"

"""Twisted client library to handle Http Streaming.

This code is an adaptation of https://github.com/fiorix/twisted-twitter-stream
"""

import base64
from twisted.protocols import basic
from twisted.internet import protocol, ssl
from twisted.python import log

from urlparse import urlparse

try:
    import simplejson as _json
except ImportError:
    try:
        import json as _json
    except ImportError:
        raise RuntimeError("A JSON parser is required, e.g., simplejson at "
                           "http://pypi.python.org/pypi/simplejson/")


class MessageReceiver(object):
    def connectionMade(self):
        pass

    def connectionFailed(self, why):
        pass

    def messageReceived(self, message):
        raise NotImplementedError

    def _registerProtocol(self, protocol):
        self._streamProtocol = protocol

    def disconnect(self):
        if hasattr(self, "_streamProtocol"):
            self._streamProtocol.factory.continueTrying = 0
            self._streamProtocol.transport.loseConnection()
        else:
            raise RuntimeError("not connected")


class HttpStreamProtocol(basic.LineReceiver):
    delimiter = "\r"

    def __init__(self):
        self.in_header = True
        self.header_data = []
        self.status_data = ""
        self.status_size = None

    def connectionMade(self):
        self.transport.write(self.factory.header)
        self.factory.consumer._registerProtocol(self)

    def lineReceived(self, line):
        while self.in_header:
            if line:
                self.header_data.append(line)
            else:
                http, status, message = self.header_data[0].split(" ", 2)
                status = int(status)
                if status == 200:
                    self.factory.consumer.connectionMade()
                else:
                    self.factory.continueTrying = 0
                    self.transport.loseConnection()
                    self.factory.consumer.connectionFailed(RuntimeError(status, message))

                self.in_header = False
            break
        else:
            #log.msg('!(%s): %s!' % (len(line), line))
            if '{' in line:
                try:
                    message = _json.loads(unicode(line.strip(), 'UTF-8'))
                except:
                    log.err()
                    log.err('%s' % line)
                self.factory.consumer.messageReceived(message)

class HttpStreamFactory(protocol.ReconnectingClientFactory):
    maxDelay = 120
    protocol = HttpStreamProtocol

    def __init__(self, consumer):
        if isinstance(consumer, MessageReceiver):
            self.consumer = consumer
        else:
            raise TypeError("consumer should be an instance of httpstream.MessageReceiver")


    def make_header(self, username, password, method, url, postdata=""):
        auth = base64.encodestring("%s:%s" % (username, password)).strip()
        header = [
            "%s %s HTTP/1.1" % (method, url.path),
            "Authorization: Basic %s" % auth,
            "User-Agent: LuckyStrike",
            "Host: %s" % url.netloc,
            "Accept: 'application/json",
        ]

        if method == "GET":
            self.header = "\r\n".join(header) + "\r\n\r\n"

        elif method == "POST":
            header += [
                "Content-Type: application/x-www-form-urlencoded",
                "Content-Length: %d" % len(postdata),
            ]
            self.header = "\r\n".join(header) + "\r\n\r\n" + postdata

def stream(reactor, url, consumer, username=None, password=None):
    url = urlparse(url)

    tw = HttpStreamFactory(consumer)
    tw.make_header(username, password, "GET", url)

    if url.scheme == "https":
        return (tw, reactor.connectSSL(url.netloc, 443, tw, ssl.ClientContextFactory()))
    else:
        return (tw, reactor.connectTCP(url.netloc, 80, tw))



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import luckystrike

setup(
    name='luckystrike',
    author='Matt Michie',
    maintainer_email='mmichie@gmail.com',
    version=luckystrike.version,
    url='https://github.com/mmichie/luckystrike',
    download_url = 'https://github.com/mmichie/luckystrike/tarball/0.1',
    keywords = ['campfire', 'irc', 'bridge'],
    license='The MIT License (MIT)',
    description='IRC to Campfire Bridge',
    long_description=open('README.md').read(),
    packages=['luckystrike'],
    scripts=['lstrike.py'],
    classifiers=[
        'Topic :: Communications :: Chat :: Internet Relay Chat',
        'Programming Language :: Python :: 2.7',
        'License :: OSI Approved :: MIT License',
    ],
    install_requires=[
        'pinder >= 1.0',
        'twisted >= 13.2.0',
        'pycrypto >= 2.6.1',
        'pyopenssl >= 0.14',
        'pyasn1 >= 0.1.7',
    ],
)
